"""
The MIT License (MIT)
Copyright © 2021 Walkline Wang (https://walkline.wang)
Gitee: https://gitee.com/walkline/a-batch-tool
"""

# Meta
__version__ = '0.7.11'
__version_info__ = (0, 7, 11)
__license__ = "MIT" # See LICENSE.txt
__author__ = 'Walkline Wang'
__email__ = 'walkline@163.com'
